// This header file holds the dispenser class.  
// This object acts as a communication link between the
// the UI and the prediction library.
// During prediction testing it will collect all the 
// required information about testing parameters and
// will dispense user motion records to a given
// predictor during test runs. (It acts like a tracking 
// device, spitting out records).
// It also will collect summary statistics and other info and
// send it to the UI for display.

#ifndef DISPENSER_H
#define DISPENSER_H

#include "types.H"
#include "list.H"
#include "math3.H"
#include "gauss.H"

class GlobalWidget;
class PredictionRec;
class ParamBundle;
class ErrorNum;
class Results;
class PredictorHolder;
class Pose;
class SignalAnalyzer;

class Dispenser 
{
  public:
   Dispenser(GlobalWidget *gbw);
   virtual ~Dispenser();

   void reset();
   void collectTestInfo();
   void fillBundle();
   Pose * addNoise(Pose *curpos, int srate, double noisev); //adds random noise to the pose 
   void   createNoise(); 
   int    getLeadTime()   { return _leadtime; }
   double getPInterval()  { return _pinterval; }

   // testing methods
   PredictionRec * runTest(int leadtime, double pinterval, int srate, double noisev);
   void runSingleTest();
   void runOverSRate();
   void runOverPInt();
   void runOverErrVar();
   void runOverAll(); // test designed to collect data points for modeling

   void runParamTest();
   void localPolyTests();   
   void greyParamTest(); 
   void kalmanParamTest();
   void ciParamTest();
   void expSmoothParamTest();

   void findOptimalParams();
   void runMCSimulation(FILE *fpp, FILE *fpq);
 
  protected:
   GlobalWidget    *_gbw;
   int              _leadtime;
   double           _pinterval;
   int              _srate;
   double           _noisev;
   PredictorType    _ptype;
   Results         *_res;  
   ParamBundle     *_bundle;
   PredictorHolder *_pholder;

   GaussWhiteNoise *_posnoise;
   GaussWhiteNoise *_quatnoise;

   // hold optimal algorithmic params
   double _alphap;
   double _alphaq;

};

#endif
