#ifndef MATH3_HAS_BEEN_INCLUDED
#define MATH3_HAS_BEEN_INCLUDED

#include <math.h>
#include <limits.h>
#include <iostream.h>

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#ifndef M_HALF_PI
#define M_HALF_PI (M_PI / 2.0)
#endif

#define CVec3i  const Vec3i
#define CVec3d  const Vec3d
#define CPt3d   const Pt3d
#define CMat3x3 const Mat3x3
#define CQuat   const Quat

class Vec3i;
class Vec3d;
class Quat;
class Mat3x3;


class Vec3i {
 protected:
   int v_[3];
 public:
   Vec3i()          { v_[0]=0;  v_[1]=0;  v_[2]=0; }
   Vec3i(CVec3i &v) { v_[0] = v.v_[0];  v_[1]=v.v_[1];  v_[2]=v.v_[2]; }
   Vec3i(int x)               { v_[0]=x;  v_[1]=x;  v_[2]=x; }
   Vec3i(int x, int y, int z) { v_[0]=x;  v_[1]=y;  v_[2]=z; }
//   ~Vec3i() {}

   operator const int * () const { return v_; }

   int  operator [] (int i) const { return v_[i]; }
   int &operator [] (int i)       { return v_[i]; }

   int      operator == (CVec3i &v) {
      return v_[0] == v.v_[0] && v_[1] == v.v_[1] && v_[2] == v.v_[2];
   }
   int      operator != (CVec3i &v) {
      return v_[0] != v.v_[0] || v_[1] != v.v_[1] || v_[2] != v.v_[2];
   }
   CVec3i & operator =  (CVec3i &v) {
      v_[0] = v[0];  v_[1] = v[1];  v_[2] = v[2];
      return *this;
   }
   CVec3i & operator += (CVec3i &v) {
      v_[0] += v[0];  v_[1] += v[1];  v_[2] += v[2];
      return *this;
   }
   CVec3i & operator -= (CVec3i &v) {
      v_[0] -= v[0];  v_[1] -= v[1];  v_[2] -= v[2];
      return *this;
   }
   CVec3i & operator *= (double s) {
      v_[0] = (int)(v_[0] * s);
      v_[1] = (int)(v_[1] * s);
      v_[2] = (int)(v_[2] * s);
      return *this;
   }
   CVec3i & operator /= (double s) {
      v_[0] = (int)(v_[0] / s);
      v_[1] = (int)(v_[1] / s);
      v_[2] = (int)(v_[2] / s);
      return *this;
   }

   Vec3i operator - () {
      return Vec3i(-v_[0], -v_[1], -v_[2]);
   }
   Vec3i operator + (CVec3i &v) {
      return Vec3i(v_[0]+v.v_[0], v_[1]+v.v_[1], v_[2]+v.v_[2]);
   }
   Vec3i operator - (CVec3i &v) {
      return Vec3i(v_[0]-v.v_[0], v_[1]-v.v_[1], v_[2]-v.v_[2]);
   }
};


class Vec3d {
 protected:
   double v_[3];
 public:
   Vec3d()                              {}// v_[0]=0;  v_[1]=0;  v_[2]=0; }
   Vec3d(CVec3d &v) { v_[0] = v.v_[0];  v_[1]=v.v_[1];  v_[2]=v.v_[2]; }
   Vec3d(double x)                      { v_[0]=x;  v_[1]=x;  v_[2]=x; }
   Vec3d(double x, double y, double z)  { v_[0]=x;  v_[1]=y;  v_[2]=z; }
   Vec3d(double v[3])    { v_[0] = v[0];  v_[1] = v[1];  v_[2] = v[2]; }
//   ~Vec3d() {}

   operator const double * () const { return v_; }

   double  operator [] (int i) const { return v_[i]; }
   double &operator [] (int i)       { return v_[i]; }

   int      operator == (CVec3d &);
   int      operator != (CVec3d &);
   CVec3d & operator =  (CVec3d &);
   CVec3d & operator += (CVec3d &);
   CVec3d & operator -= (CVec3d &);
   CVec3d & operator *= (double);
   CVec3d & operator /= (double);
   Vec3d   operator *  (double s) const {
      return Vec3d(v_[0]*s, v_[1]*s, v_[2]*s); }
   operator Vec3i() {
      return Vec3i(int(v_[0]*INT_MAX), int(v_[1]*INT_MAX), int(v_[2]*INT_MAX));
   }

   void set(double x, double y, double z) { v_[0]=x;  v_[1]=x;  v_[2]=x; }

   double dot(CVec3d &v) const {
      return v_[0]*v.v_[0] + v_[1]*v.v_[1] + v_[2]*v.v_[2]; }

   CVec3d & truncate( double max );

   static const Vec3d zero;
   static const Vec3d north;
   static const Vec3d south;
   static const Vec3d east;
   static const Vec3d west;
   static const Vec3d up;
   static const Vec3d down;

   static const Vec3d basis[3];
};


inline CVec3d &
Vec3d::operator += (CVec3d &t) {
   v_[0] += t[0];   v_[1] += t[1];   v_[2] += t[2];
   return *this;
}


inline CVec3d &
Vec3d::operator -= (CVec3d &t) {
   v_[0] -= t[0];   v_[1] -= t[1];   v_[2] -= t[2];
   return *this;
}


class Pt3d {
 protected:
   double v_[3];
 public:
   Pt3d()                             { v_[0]=0;  v_[1]=0;  v_[2]=0; }
   Pt3d(CPt3d &v) { v_[0] = v.v_[0];  v_[1]=v.v_[1];  v_[2]=v.v_[2]; }
   Pt3d(double x)                     { v_[0]=x;  v_[1]=x;  v_[2]=x; }
   Pt3d(double x, double y, double z) { v_[0]=x;  v_[1]=y;  v_[2]=z; }
   Pt3d(double v[3])   { v_[0] = v[0];  v_[1] = v[1];  v_[2] = v[2]; }
//   ~Pt3d() {}

   operator const double * () const { return v_; }

   double  operator [] (int i) const { return v_[i]; }
   double &operator [] (int i)       { return v_[i]; }

   int     operator == (CPt3d &);
   int     operator != (CPt3d &);
   CPt3d & operator =  (CPt3d &);
   CPt3d & operator += (CVec3d &);
   CPt3d & operator -= (CVec3d &);
};


inline CPt3d &
Pt3d::operator += (CVec3d &t) {
   v_[0] += t[0];   v_[1] += t[1];   v_[2] += t[2];
   return *this;
}


inline CPt3d &
Pt3d::operator -= (CVec3d &t) {
   v_[0] -= t[0];   v_[1] -= t[1];   v_[2] -= t[2];
   return *this;
}


class Quat {
 public:
   Quat(): s_(1), v_(0, 0, 0) {}
   Quat(CQuat &q): s_(q.s_), v_(q.v_) {}
   Quat(double s, CVec3d &v): s_(s), v_(v) {}
   Quat(double x, double y, double z, double s): s_(s), v_(x, y, z) {}
   Quat(CMat3x3 &m);
   Quat(CVec3d &axis, double angle);
//   ~Quat() {}

   double & S()       { return s_; }
   Vec3d  & V()       { return v_; }
   double   S() const { return s_; }
   CVec3d & V() const { return v_; }

   CQuat & operator += (CQuat &q) { s_ += q.s_;  v_ += q.v_;  return *this; }
   CQuat & operator -= (CQuat &q) { s_ -= q.s_;  v_ -= q.v_;  return *this; }
   CQuat & operator =  (CQuat &q);
 
  double length() {return sqrt((v_[0]*v_[0])+(v_[1]*v_[1])+(v_[2]*v_[2])+(s_*s_)); }  
  
  void set(double x, double y, double z, double w) { s_ = w; v_.set(x,y,z); }
  
  void printQuat() { cerr << "[("<< v_[0]<< ", " << v_[1] << ", " << v_[2] << "), " 
                          << s_ << "]" << endl; }
                          

 protected:
   double s_;
   Vec3d  v_;
};


class Mat3x3 {
 public:
   Mat3x3() {
      double *p = &m_[0][0];
      *p++ = 1;   *p++ = 0;   *p++ = 0;
      *p++ = 0;   *p++ = 1;   *p++ = 0;
      *p++ = 0;   *p++ = 0;   *p   = 1;
   }
   Mat3x3(CMat3x3 &);
   Mat3x3(double,double,double,double,double,double,double,double,double);
   Mat3x3(CQuat &q);
//   ~Mat3x3() {}

   double & operator () (int i, int j)       { return m_[i][j]; }
   double   operator () (int i, int j) const { return m_[i][j]; }

   Mat3x3 &Normalize();
 protected:
   double m_[3][3];
};


class MatRotX : public Mat3x3 {
 public:
   MatRotX(double a) {
      const double c = cos(a), s = sin(a);
      double *p = &m_[0][0];
      *p++ =  1;   *p++ =  0;   *p++ =  0;
      *p++ =  0;   *p++ =  c;   *p++ = -s;
      *p++ =  0;   *p++ =  s;   *p   =  c;
   }
};


class MatRotY : public Mat3x3 {
 public:
   MatRotY(double a) {
      const double c = cos(a), s = sin(a);
      double *p = &m_[0][0];
      *p++ =  c;   *p++ =  0;   *p++ =  s;
      *p++ =  0;   *p++ =  1;   *p++ =  0;
      *p++ = -s;   *p++ =  0;   *p   =  c;
   }
};


class MatRotZ : public Mat3x3 {
 public:
   MatRotZ(double a) {
      const double c = cos(a), s = sin(a);
      double *p = &m_[0][0];
      *p++ =  c;   *p++ = -s;   *p++ =  0;
      *p++ =  s;   *p++ =  c;   *p++ =  0;
      *p++ =  0;   *p++ =  0;   *p   =  1;
   }
};


class MatRotAxis : public Mat3x3 {
 public:
   MatRotAxis(Vec3d v, double a) {
      const double
         s = sin(a), c = cos(a), c_1 = 1-c,
         xx = v[0]*v[0]*c_1, xy = v[0]*v[1]*c_1, xz = v[0]*v[2]*c_1,
         yy = v[1]*v[1]*c_1, yz = v[1]*v[2]*c_1, zz = v[2]*v[2]*c_1,
         xs = v[0]*s, ys = v[1]*s, zs = v[2]*s;
      double *p = &m_[0][0];
      *p++ = xx+c;   *p++ = xy-zs;  *p++ = xz+ys;
      *p++ = xy+zs;  *p++ = yy+c;   *p++ = yz-xs;
      *p++ = xz-ys;  *p++ = yz+xs;  *p   = zz+c;
   }
};


/* ---- inline functions ---- */

inline Vec3i
operator * (double s, CVec3i &a) {
   return Vec3i((int)(s*a[0]), (int)(s*a[1]), (int)(s*a[2]));
}


inline Vec3i
operator * (CVec3i &a, double s) {
   return Vec3i((int)(s*a[0]), (int)(s*a[1]), (int)(s*a[2]));
}


inline Vec3i
operator / (CVec3i &a, double s) {
   return Vec3i((int)(a[0]/s), (int)(a[1]/s), (int)(a[2]/s));
}


inline Quat::Quat (CMat3x3 &m) {
   double tr = m(0,0) + m(1,1) + m(2,2), s;
   if (tr >= 0) {
      s = sqrt(tr + 1);
      s_ = 0.5 * s;
      s = 0.5 / s;
      v_[0] = m(2,1) - m(1,2) * s;
      v_[1] = m(0,2) - m(2,0) * s;
      v_[2] = m(1,0) - m(0,1) * s;
   } else {
      int i = 0;
      if (m(1,1) > m(0,0)) i = 1;
      if (m(2,2) > m(i,i)) i = 2;
      switch (i) {
       case 0:
         s = sqrt((m(0,0) - (m(1,1) + m(2,2))) + 1);
         v_[0] = 0.5 * s;
         s = 0.5 / s;
         v_[1] = (m(0,1) + m(1,0)) * s;
         v_[2] = (m(2,0) + m(0,2)) * s;
         s_    = (m(2,1) - m(1,2)) * s;
         break;
       case 1:
         s = sqrt((m(1,1) - (m(2,2) + m(0,0))) + 1);
         v_[1] = 0.5 * s;
         s = 0.5 / s;
         v_[2] = (m(1,2) + m(2,1)) * s;
         v_[0] = (m(0,1) + m(1,0)) * s;
         s_    = (m(0,2) - m(2,0)) * s;
         break;
       case 2:
         s = sqrt((m(2,2) - (m(0,0) + m(1,1))) + 1);
         v_[2] = 0.5 * s;
         s = 0.5 / s;
         v_[0] = (m(2,0) + m(0,2)) * s;
         v_[1] = (m(1,2) + m(2,1)) * s;
         s_    = (m(1,0) - m(0,1)) * s;
         break;
      }
   }
}


inline Mat3x3::Mat3x3 (CMat3x3 &a) {
   double *p = (double *)(&m_[0][0]), *q = (double *)(&a.m_[0][0]);
   *p++ = *q++;   *p++ = *q++;   *p++ = *q++;
   *p++ = *q++;   *p++ = *q++;   *p++ = *q++;
   *p++ = *q++;   *p++ = *q++;   *p   = *q;
}


inline
Mat3x3::Mat3x3(double a, double b, double c,
               double d, double e, double f,
               double g, double h, double i) {
   double *p = &m_[0][0];
   *p++ = a;   *p++ = b;   *p++ = c;
   *p++ = d;   *p++ = e;   *p++ = f;
   *p++ = g;   *p++ = h;   *p   = i;
}


inline
Mat3x3::Mat3x3(CQuat &q) {
   const double s = q.S(), x = q.V()[0], y = q.V()[1], z = q.V()[2],
      sx2 = s*x*2, sy2 = s*y*2, sz2 = s*z*2,
      xx2 = x*x*2, xy2 = x*y*2, xz2 = x*z*2,
      yy2 = y*y*2, yz2 = y*z*2, zz2 = z*z*2;
   double *p = &m_[0][0];
   *p++ = 1-yy2-zz2;  *p++ =   xy2-sz2;  *p++ =   xz2+sy2;
   *p++ =   xy2+sz2;  *p++ = 1-xx2-zz2;  *p++ =   yz2-sx2,
   *p++ =   xz2-sy2;  *p++ =   yz2+sx2;  *p   = 1-xx2-yy2;
}


inline void
QuatToGLmatrix(double v[16], CQuat &q) {
   const double s = q.S(), x = q.V()[0], y = q.V()[1], z = q.V()[2],
      sx2 = s*x*2, sy2 = s*y*2, sz2 = s*z*2,
      xx2 = x*x*2, xy2 = x*y*2, xz2 = x*z*2,
      yy2 = y*y*2, yz2 = y*z*2, zz2 = z*z*2;
   double *p = &v[0];
   *p++ = 1-yy2-zz2;  *p++ =   xy2+sz2;  *p++ =   xz2-sy2;  *p++ = 0.0;
   *p++ =   xy2-sz2;  *p++ = 1-xx2-zz2;  *p++ =   yz2+sx2,  *p++ = 0.0;
   *p++ =   xz2+sy2;  *p++ =   yz2-sx2;  *p++ = 1-xx2-yy2;  *p++ = 0.0;
   *p++ = 0.0;        *p++ = 0.0;        *p++ = 0.0;        *p   = 1.0;
}


inline Mat3x3
Star(Vec3d &a) {
   return Mat3x3(   0,  -a[2],  a[1],
                  a[2],    0,  -a[0],
                 -a[1],  a[0],    0);
}


inline Mat3x3
operator * (CMat3x3 &a, CMat3x3 &b) {
   Mat3x3 r;
   for (int i=0; i<3; i++)
      for (int j=0; j<3; j++)
         r(i,j) = a(i,0) * b(0,j) + a(i,1) * b(1,j) + a(i,2) * b(2,j);
   return r;
}


inline Mat3x3
Transpose(CMat3x3 &m) {
   return Mat3x3(m(0,0), m(1,0), m(2,0),
                 m(0,1), m(1,1), m(2,1),
                 m(0,2), m(1,2), m(2,2));
}


inline Vec3d
operator * (CMat3x3 &m, CVec3d &v) {
   return Vec3d(m(0,0)*v[0] + m(0,1)*v[1] + m(0,2)*v[2],
               m(1,0)*v[0] + m(1,1)*v[1] + m(1,2)*v[2],
               m(2,0)*v[0] + m(2,1)*v[1] + m(2,2)*v[2]);
}


// multiplies a vector on the right side of a matrix
//  equiv to mult by the inverse if m is a rotation matrix
inline Vec3d
operator * (CVec3d &v, CMat3x3 &m) {
   return Vec3d(v[0]*m(0,0) + v[1]*m(1,0) + v[2]*m(2,0),
               v[0]*m(0,1) + v[1]*m(1,1) + v[2]*m(2,1),
               v[0]*m(0,2) + v[1]*m(1,2) + v[2]*m(2,2));
}


inline CVec3d &
Vec3d::operator =  (CVec3d &a) {
   if (&a != this) {
      v_[0] = a.v_[0];  v_[1] = a.v_[1];  v_[2] = a.v_[2];
   }
   return *this;
}


inline int
Vec3d::operator == (CVec3d &a) {
   return v_[0] == a.v_[0] && v_[1] == a.v_[1] && v_[2] == a.v_[2];
}


inline int
Vec3d::operator != (CVec3d &a) {
   return v_[0] != a.v_[0] || v_[1] != a.v_[1] || v_[2] != a.v_[2];
}


inline CVec3d &
Vec3d::operator *= (double s) {
   v_[0] *= s;   v_[1] *= s;   v_[2] *= s;
   return *this;
}


inline CVec3d &
Vec3d::operator /= (double s) {
   v_[0] /= s;   v_[1] /= s;   v_[2] /= s;
   return *this;
}


inline Vec3d
operator + (CVec3d &a, CVec3d &b) {
   return Vec3d(a[0]+b[0], a[1]+b[1], a[2]+b[2]);
}


inline Vec3d
operator - (CVec3d &a, CVec3d &b) {
   return Vec3d(a[0]-b[0], a[1]-b[1], a[2]-b[2]);
}


inline Vec3d
operator - (CVec3d &a) {
   return Vec3d(-a[0], -a[1], -a[2]);
}


inline double
dot (CVec3d &a, CVec3d &b) {
//   return double(a[0] * b[0] + a[1] * b[1] + a[2] * b[2]);
   return a.dot(b);
}


inline Vec3d
cross (CVec3d &a, CVec3d &b) {
   return Vec3d(a[1]*b[2]-a[2]*b[1], a[2]*b[0]-a[0]*b[2], a[0]*b[1]-a[1]*b[0]);
}


inline Vec3d
operator * (double s, CVec3d &a) {
   return a*s; //Vec3d(s*a[0], s*a[1], s*a[2]);
}

/////////////////////////////////////////////////////////////
// these operations only makes sense if we are trying to do component wise
// error analysis or other component wise operations
inline Vec3d
operator * (CVec3d &a, CVec3d &b) {
   return Vec3d(a[0]*b[0], a[1]*b[1], a[2]*b[2]);
}

inline Quat
specialmultiply(CQuat &a, CQuat &b)
{
 return Quat(a.V()[0]*b.V()[0], a.V()[1]*b.V()[1], a.V()[2]*b.V()[2], a.S()*b.S());
}

inline Vec3d
mysqrt (CVec3d &a) {
  return Vec3d(sqrt(a[0]), sqrt(a[1]), sqrt(a[2]));
}

inline Quat
mysqrt (CQuat &a) {
  return Quat(sqrt(a.V()[0]), sqrt(a.V()[1]), sqrt(a.V()[2]), sqrt(a.S()));
}


inline Vec3d
myfabs(CVec3d &a, CVec3d &b)
{
 return Vec3d(fabs(a[0]-b[0]), fabs(a[1]-b[1]), fabs(a[2]-b[2]));
}

inline Quat
myfabs(CQuat &a, CQuat &b)
{
 return Quat(fabs(a.V()[0]-b.V()[0]), fabs(a.V()[1]-b.V()[1]), fabs(a.V()[2]-b.V()[2]),
             fabs(a.S()-b.S()));
}

inline Quat
operator / (CQuat &a, double s) {
  return Quat(a.V()[0] / s, a.V()[1] / s, a.V()[2] / s, a.S() / s);
}


/////////////////////////////////////////////////////////////

inline Vec3d
operator / (CVec3d &a, double s) {
   return Vec3d(a[0]/s, a[1]/s, a[2]/s);
}


inline double
length2 (CVec3d &a) {
   return dot(a, a);
}


inline double
length (CVec3d &a) {
   return sqrt(dot(a, a));
}


inline double
distance1 (CVec3d &a, CVec3d &b) {
   Vec3d v = b - a;
   return sqrt(dot(v, v));
}


inline double
distance2 (CVec3d &a, CVec3d &b) {
   Vec3d v = b - a;
   return dot(v, v);
}


inline double
distance1 (CPt3d &a, CPt3d &b) {
   Vec3d v = b - a;
   return sqrt(dot(v, v));
}


inline double
distance2 (CPt3d &a, CPt3d &b) {
   Vec3d v = b - a;
   return dot(v, v);
}


inline Vec3d
normalize (CVec3d &a) {
   return Vec3d(a / length(a));
}



inline CVec3d &
Vec3d::truncate( double max )
{
   double len = length( *this );
  
   double ratio = len / max;
  
   if ( ratio > 1.0 ) {
      *this /= ratio;
   }
  
   return *this;
}



inline CPt3d &
Pt3d::operator =  (CPt3d &a) {
   if (&a != this) {
      v_[0] = a.v_[0];  v_[1] = a.v_[1];  v_[2] = a.v_[2];
   }
   return *this;
}


inline int
Pt3d::operator == (CPt3d &a) {
   return v_[0] == a.v_[0] && v_[1] == a.v_[1] && v_[2] == a.v_[2];
}


inline int
Pt3d::operator != (CPt3d &a) {
   return v_[0] != a.v_[0] || v_[1] != a.v_[1] || v_[2] != a.v_[2];
}


inline Pt3d
operator + (CPt3d &a, CVec3d &b) {
   return Pt3d(a[0]+b[0], a[1]+b[1], a[2]+b[2]);
}


inline Pt3d
operator - (CPt3d &a, CVec3d &b) {
   return Pt3d(a[0]-b[0], a[1]-b[1], a[2]-b[2]);
}


inline Vec3d
operator - (CPt3d &a, CPt3d &b) {
   return Vec3d(a[0]-b[0], a[1]-b[1], a[2]-b[2]);
}


inline Pt3d
operator - (CPt3d &a) {
   return Pt3d(-a[0], -a[1], -a[2]);
}


inline Quat
normalize (CQuat &q) {
   double l = 1.0 / sqrt(q.S() * q.S() + dot(q.V(), q.V()));
   return Quat(q.S()*l, q.V()*l);
}

inline CQuat &
Quat::operator = (CQuat &q) {
  if (&q != this) {
    v_[0] = q.V()[0]; v_[1] = q.V()[1]; v_[2] = q.V()[2]; s_ = q.S();
  }  
  return *this;
}

inline Quat
operator * (CQuat &q1, CQuat &q2) {
   return Quat(q1.S() * q2.S() - dot(q1.V(), q2.V()),
               q1.S() * q2.V() + q2.S() * q1.V() + cross(q1.V(), q2.V()));
}


inline Quat
operator * (CQuat &q, CVec3d &v) {
   return Quat(-dot(q.V(), v), q.S() * v + cross(q.V(), v));
}


inline Quat
operator * (CVec3d &v, CQuat &q) {
   return Quat(-dot(v, q.V()), q.S() * v + cross(v, q.V()));
}


inline Quat
operator * (double s, CQuat &q) {
   return Quat(s*q.S(), s*q.V());
}


inline Quat
operator + (CQuat &q1, CQuat &q2) {
   return Quat(q1.S()+q2.S(), q1.V()+q2.V());
}

inline Quat
operator - (CQuat &q1, CQuat &q2) {
   return Quat(q1.S()-q2.S(), q1.V()-q2.V());
}

inline Quat
inverse (CQuat &q) {
   double mag2 = q.S() * q.S() + dot(q.V(), q.V());
   return Quat(q.S()/mag2, -q.V()/mag2);
}


inline Quat
conjugate (CQuat &q) {
   return Quat(q.S(), -q.V());
}


// for unit quaternions only
inline Vec3d transform (CQuat &q, CVec3d &v) {
//   return (q * v * conjugate(q)).V();
   return Vec3d( (q.S()*q.S() - dot(q.V(), q.V())) * v +
                 2.0 * dot(v, q.V()) * q.V() +
                 2.0 * q.S() * cross(q.V(), v) );
}


// for unit quaternions only
inline Pt3d transform (CQuat &q, CPt3d &p) {
   Vec3d v(p[0], p[1], p[2]);
   v = (q * v * conjugate(q)).V();
   return Pt3d(v[0], v[1], v[2]);
}


inline Mat3x3 &
Mat3x3::Normalize() {
   const Vec3d z(normalize(Vec3d(m_[2][0], m_[2][1], m_[2][2])));
   const Vec3d x(normalize(cross(Vec3d(m_[1][0], m_[1][1], m_[1][2]), z)));
   const Vec3d y(cross(z, x));
   double *p = &m_[0][0];
   *p++ = x[0];  *p++ = x[1];  *p++ = x[2];
   *p++ = y[0];  *p++ = y[1];  *p++ = y[2];
   *p++ = z[0];  *p++ = z[1];  *p   = z[2];
   return *this;
}


class MatOrient : public Mat3x3 {
 public:
   MatOrient(Vec3d at, Vec3d up) {
      Vec3d z(-normalize(at));
      Vec3d x(normalize(cross(up, z)));
      Vec3d y(cross(z, x));
      double *p = &m_[0][0];
      *p++ = x[0];  *p++ = y[0];  *p++ = z[0];
      *p++ = x[1];  *p++ = y[1];  *p++ = z[1];
      *p++ = x[2];  *p++ = y[2];  *p   = z[2];
   }
};


inline Quat::Quat(CVec3d &axis, double angle):
   s_(cos(0.5*angle)), v_(sin(0.5*angle)*axis) {}


#define EPSILON 1.0E-6
#define CUTOFF_ANGLE M_PI

inline 
Vec3d logMap(CQuat &q1)
{
   double w = q1.S();
   Vec3d tmpv = q1.V();

   if (w == 1)
      return Vec3d(0.0,0.0,0.0);
   else
      return (2.0 * acos(w) / length(tmpv)) * tmpv;
   
}


inline
Quat expMap(CVec3d &v)
{
   double theta = length(v);
   Vec3d oldv = v;

   // reparameterize if required
   /*  if (theta > CUTOFF_ANGLE) {
      double scl = theta; 
      if (theta > 2 *M_PI) {
	 theta = fmod(theta, 2*M_PI);
	 scl = theta / scl;
	 oldv = scl * v;
      }
      if (theta > CUTOFF_ANGLE) {
	 scl = theta;
	 theta = 2.0 * M_PI - theta;
	 scl = 1.0 - 2.0 *M_PI / scl;
	 oldv = scl * v;
      }
   }*/
   
   double w = cos(0.5 * theta);
   Vec3d vv(0.0,0.0,0.0);
   if (theta < EPSILON)  // do taylor series
      vv = (0.5 + (theta*theta) / 48.0) * oldv;
   else
      vv = (sin(0.5 * theta) / theta) * oldv;
   
   return Quat(w, vv);
}


// code adapted from Graphics Gems III
inline 
Quat slerp(CQuat &q1, CQuat &q2, double alpha) {

   double beta = 0.0;
   double theta = 0.0;
   double sint = 0.0;
   double cost = 0.0;
   int bflip = 0;

   cost = q1.V()[0]*q2.V()[0] + q1.V()[1]*q2.V()[1] + q1.V()[2]*q2.V()[2] +
          q1.S()*q2.S();

   // special cases
   if (cost < 0.0) {
     cost = -cost;
     bflip = 1;
   }
   else
     bflip = 0;

   if ((1.0 - cost) < EPSILON) 
      beta = 1 - alpha;
   else {
     theta = acos(cost);
     sint = sin(theta);
     beta = sin((1.0 - alpha)*theta) / sint;
     alpha = sin(alpha* theta) / sint;
   }

   if (bflip) 
     alpha = -alpha;
   
   // interpolate
   return Quat(beta * q1.V()[0] + alpha * q2.V()[0],
               beta * q1.V()[1] + alpha * q2.V()[1],
               beta * q1.V()[2] + alpha * q2.V()[2],
               beta * q1.S()    + alpha * q2.S()); 


}

// complex number class

  
class Complex
{
 private:
   double re;
   double im;
  
 public:
   Complex(float r,float i) {re = r; im = i;}
   Complex(float r) {re = r;im = 0.0;}
   Complex() {re = 0.0; im = 0,0; }
   ~Complex() {};
  
   double magnitude()         // calculate magnitude
   { sqrt(re*re + im*im); }    

   Complex conjg() 
   { return Complex(re,-im); }
    
   double real() {return re;} // return real part
   double imag() {return im;} // return imaginary part
   
   Complex operator+(Complex b)
   {return Complex(re + b.re,im + b.im);}
  
   Complex operator+=(Complex b)
   {re += b.re; im += b.im; }
  
   Complex operator=(Complex b)
   {re = b.re;im = b.im; return *this;}
 
   Complex operator/(Complex b)
   {return Complex((re*b.re + im*b.im)/(b.re*b.re + b.im*b.im),
                   (im*b.re - re*b.im)/(b.re*b.re + b.im*b.im)); }

   Complex operator*(Complex b)
   {return Complex(re*b.re - im*b.im, re*b.im + im*b.re); }
 
};
  
  



#endif  // MATH3_HAS_BEEN_INCLUDED

