#include "types.H"
#include "ci.H"
#include "ciPVSM.H"
#include "ciPVMMV.H"

CIPredictor::CIPredictor()
{
   // create the different ci filter predictors
   _cipvsm = new CIPVSM();
   _cipvmmv = new CIPVMMV();

}

CIPredictor::~CIPredictor()
{
   delete _cipvsm;
   delete _cipvmmv;
}

void
CIPredictor::prepare()
{
  // get some things we need from the bundle and send them to
  // the appropriate filter
  double spd = (double)_paramb->getCISpecD();
  double delta = 1.0 / _paramb->getCurSRate();
  double mnoisep = _paramb->getPosVar();
  double mnoiseq = _paramb->getQuatVar();

  switch (_paramb->getCIType()) {
    case tCIPVSM:
      _cipvsm->initialize(spd, delta, mnoisep, mnoiseq);
      break;
    case tCIPVMMV:
      _cipvmmv->initialize(spd, delta, mnoisep, mnoiseq);
      break;
    default: 
    cerr << "No Covariance Intersection Predictor Selected!!" << endl;
  }
}

Pose *
CIPredictor::predict(Pose *userpos, double pinterval)
{
  switch (_paramb->getCIType()) {
    case tCIPVSM:
      return _cipvsm->filter(userpos, pinterval);
      break;
    case tCIPVMMV:
      return _cipvmmv->filter(userpos, pinterval);
      break;
    default:
      cerr << "No Predictor Selected!!!" << endl;
      // for testing, just make everythin constant
      Pose *tmp = new Pose(userpos->_tstamp+pinterval, Vec3d(0.0, 0.0, 0.0), Quat(0.0,0.0,0.0,1.0));
      return tmp;
  }
}



