// functions in this these objects do generic kalman,extended, and
// unscented kalman filtering


#include "kalmanbase.H"

#define alpha 1
#define beta 0
#define kappa 0

// standard kalman filtering
void KalmanBase::kalmanFilter(Pose *userpos, int zNum, int iNum, KALMANMODEL mod)
{
 for (int i = 0; i < 3; ++i) {

   // do the time update step
   MMatrix pstate = _fundMat * _state[i];

   // do the a priori extimate of the the process covariance matrix
   MMatrix pcov = _fundMat * (_PP[i] * transp(_fundMat)) + _QQ;

   // compute the Kalman Gain

   MMatrix tmp = _HH * (pcov * transp(_HH)) + _RR;

   MMatrix KG = pcov * (transp(_HH) * minverse(tmp));

   // do the measurment update step
   MMatrix Z(zNum,1);
   switch(mod) {
    case kPVSM:
    case kPVASM: 
      Z(0,0) = userpos->_pos[i];
      break;
    case kPVMMV:
    case kPVAMMV:
      Z(0,0) = userpos->_pos[i];
      Z(1,0) = userpos->_posvel[i]; 
      break;
    case kPVAMMA:
      Z(0,0) = userpos->_pos[i];
      Z(1,0) = userpos->_posacc[i];
      break;
    case kPVAMMVA:
      Z(0,0) = userpos->_pos[i];
      Z(1,0) = userpos->_posvel[i]; 
      Z(2,0) = userpos->_posacc[i];
      break;
    default: ;
   }

   _state[i] = pstate + KG * (Z - (_HH * pstate));

   // update the covariance matrix
   MMatrix I(iNum,iNum);
   I.setIdentity();
   
   _PP[i] = (I - (KG * _HH)) * pcov;   

   //    if (i == 0)
   //    cerr << " " << _state[i](0,0) << " " << sqrt(_PP[i](0,0)) << endl;
 }
}



// extended kalman filtering
void 
ExtendedKalmanBase::extendedKalmanFilter(Pose *userpos, int zNum, int iNum,
                                         KALMANMODEL mod)
{
   // do the time update step using _edt and _Estate
   MMatrix epstate = _up->RK4(_Estate, _edt, mod);

  // update the fundamental matrix, Q matrix, and the H matrix
  _EfundMat = _up->updateFundMat(epstate, _edt, mod);
  _EQ = _up->updateQMat(epstate, _edt, _spd, mod);
  _EH = _up->updateHMat(epstate, mod);

  // do the a priori estimate of the covariance matrix
  MMatrix epcov = _EfundMat * (_EP * transp(_EfundMat)) + _EQ;

   // compute the Kalman Gain
   MMatrix etmp = _EH * (epcov * transp(_EH)) + _ER;
   MMatrix EKG = epcov * (transp(_EH) * minverse2(etmp));
   // do the mearurment update
   MMatrix EZ(zNum,1);
   switch(mod) {
     case kPVSM:
     case kPVASM:    
       for (int i = 0; i < 3; ++i)
	  EZ(i,0) = userpos->_orient.V()[i];
       EZ(3,0) = userpos->_orient.S();
       break;
     case kPVMMV:
     case kPVAMMV:
       for (int i = 0; i < 3; ++i) {
	  EZ(i,0) = userpos->_orient.V()[i];
	  EZ(i+4,0) = userpos->_angvel.V()[i];
       }
       EZ(3,0) = userpos->_orient.S();
       break;
     case kPVAMMA:
       for (int i = 0; i < 3; ++i) {
	  EZ(i,0) = userpos->_orient.V()[i];
	  EZ(i+4,0) = userpos->_angacc.V()[i];
       }
       EZ(3,0) = userpos->_orient.S();
       break;
     case kPVAMMVA:
       for (int i = 0; i < 3; ++i) {
	  EZ(i,0) = userpos->_orient.V()[i];
	  EZ(i+4,0) = userpos->_angvel.V()[i];
	  EZ(i+7,0) = userpos->_angacc.V()[i];
       }
       EZ(3,0) = userpos->_orient.S();
       break;
     default: ;
   }
   _Estate = epstate + EKG * (EZ - (_EH * epstate));

   MMatrix EI(iNum,iNum);
   EI.setIdentity();
   // update covariance
   _EP = (EI - (EKG * _EH)) * epcov;
   
   // better normalize the quaternion just in case
   Quat tmpq = normalize(Quat(_Estate(0,0), _Estate(1,0), _Estate(2,0), _Estate(3,0)));

   // put it back in the state vector
   for (int i = 0; i < 3; ++i)
     _Estate(i,0) = tmpq.V()[i];
   _Estate(3,0) = tmpq.S();
   
   if (getenv("FILTER_DATA"))
       cerr << " " << _Estate(0,0) << " " << _Estate(1,0) << " " << _Estate(2,0) << " " << _Estate(3,0) << " " << _EP(0,0) << " " << _EP(1,1) << " " << _EP(2,2) << " " << _EP(3,3) << endl;
}


// unscented kalman filtering

void 
UnscentedKalmanBase::unscentedKalmanFilter(Pose *userpos, int zNum, KALMANMODEL mod)
{
// first compute square root of covariance matrix 
  // use cholesky factorization
  MMatrix sqrtP = choleskyDecomp(_UP);

  // calculate the sigma points
  MMatrix sigma1(_dim, 2 * _dim + 1);
  double coeff = sqrt(_dim + (alpha*alpha * (_dim + kappa) - _dim));
  sigma1.insertColumn(_Ustate, 0);
  for (int i = 1; i < 2*_dim + 1; ++i) 
    if (i < _dim+1) 
     sigma1.insertColumn(_Ustate + (coeff * sqrtP.getColumn(i-1)), i);
    else
     sigma1.insertColumn(_Ustate - (coeff * sqrtP.getColumn(i-_dim-1)), i);
  
   
  // now we must progagate the sigma vectors through the nonlinear equations
  for (int i = 0; i < 2 * _dim + 1; ++i)
    sigma1.insertColumn(_uup->RK4(sigma1.getColumn(i),_udt, mod), i);			
  
  // compute the a priori state estimate
  MMatrix upstate(_dim,1);
  double lambda = alpha*alpha*(_dim + kappa) - _dim;
  for (int i = 0; i < 2 * _dim + 1; ++i) 
    if (i == 0)
      upstate += (lambda / (lambda + _dim)) * sigma1.getColumn(i);
    else
      upstate += (1.0 / (2.0 * (lambda + _dim))) * sigma1.getColumn(i);

  // compute the a priori covariance matrix
  _UQ = _uup->updateQMat(upstate, _udt, _uspd, mod);
  MMatrix upcov(_dim,_dim);
  for (int i = 0; i < 2 * _dim + 1; ++i)
    if (i == 0)
     upcov += ((lambda / (_dim + lambda)) + 1 - alpha*alpha + beta) *
              (sigma1.getColumn(i) - upstate) * transp(sigma1.getColumn(i) - upstate);
    else
     upcov += (1.0 / (2.0 * (lambda + _dim))) * (sigma1.getColumn(i) - upstate) * 
              transp(sigma1.getColumn(i) - upstate);
  upcov += _UQ;

  // transform the sigma vectors into measurment space
  MMatrix sigma2(_dim - zNum, 2 * _dim + 1);
  for (int i = 0; i < 2 * _dim + 1; ++i) {
    MMatrix xcol = sigma1.getColumn(i);
    Quat qtmp = normalize(Quat(xcol(0,0), xcol(1,0), xcol(2,0), xcol(3,0)));
    for(int j = 0; j < 3; ++j)
      sigma2(j,i) = qtmp.V()[j];
    sigma2(3,i) = qtmp.S();
  }
  
  // compute the measurement estimate
  MMatrix upmeas(_dim - zNum, 1);
  for (int i = 0; i < 2 * _dim + 1; ++i) 
    if (i == 0)
      upmeas += (lambda / (lambda + _dim)) * sigma2.getColumn(i);
    else
      upmeas += (1.0 / (2.0 * (lambda + _dim))) * sigma2.getColumn(i);

  // do the measurement update step
  MMatrix pcovyy(_dim - zNum, _dim - zNum);
  for (int i = 0; i < 2 * _dim + 1; ++i)
    if (i == 0)
      pcovyy +=  ((lambda / (_dim + lambda)) + 1 - alpha*alpha + beta) *
                 (sigma2.getColumn(i) - upmeas) * transp(sigma2.getColumn(i) - upmeas);
    else
      pcovyy += (1.0 / (2.0 * (lambda + _dim))) *  (sigma2.getColumn(i) - upmeas) *
                transp(sigma2.getColumn(i) - upmeas);
  pcovyy += _UR;

  MMatrix pcovxy(_dim, _dim - zNum);
  for (int i = 0; i < 2 * _dim + 1; ++i)
    if (i == 0)
      pcovxy += ((lambda / (_dim + lambda)) + 1 - alpha*alpha + beta) *
	        (sigma1.getColumn(i) - upstate) * transp(sigma2.getColumn(i) - upmeas);
    else
      pcovxy += (1.0 / (2.0 * (lambda + _dim))) *  (sigma1.getColumn(i) - upstate) *
                transp(sigma2.getColumn(i) - upmeas);

  // compute Kalman gain 
  MMatrix UKG = pcovxy * minverse2(pcovyy);
  
  MMatrix z(_dim - zNum, 1);
  switch(mod) {
    case kPVSM: 
    case kPVASM:
      for (int i = 0; i < 3; ++i)
	 z(i,0) = userpos->_orient.V()[i];
      z(3,0) = userpos->_orient.S();
      break; 
    case kPVMMV:
    case kPVAMMV:
      for (int i = 0; i < 3; ++i) {
	 z(i,0) = userpos->_orient.V()[i];
	 z(i+4,0) = userpos->_angvel.V()[i];
      }
      z(3,0) = userpos->_orient.S();
      break;
    case kPVAMMA:   
      for (int i = 0; i < 3; ++i) {
	 z(i,0) = userpos->_orient.V()[i];
	 z(i+4,0) = userpos->_angacc.V()[i];
      }
      z(3,0) = userpos->_orient.S();
      break;
    case kPVAMMVA:  
      for (int i = 0; i < 3; ++i) {
	 z(i,0) = userpos->_orient.V()[i];
	 z(i+4,0) = userpos->_angvel.V()[i];
	 z(i+7,0) = userpos->_angacc.V()[i];
      }
      z(3,0) = userpos->_orient.S();
      break;
    default: ;
  }
 
  // update the state and covariance
  _Ustate = upstate + (UKG * (z - upmeas));

  _UP = upcov - (UKG * pcovyy * transp(UKG));

  // normalize quaternion just in case  -- may not need this
  Quat tmpq = normalize(Quat(_Ustate(0,0), _Ustate(1,0), _Ustate(2,0), _Ustate(3,0)));
   // put it back in the state vector
     for (int i = 0; i < 3; ++i)
      _Ustate(i,0) = tmpq.V()[i];
    _Ustate(3,0) = tmpq.S();
   
   //   cerr << " " << _Ustate(0,0) << " " << sqrt(_UP(0,0)) << endl;
    if (getenv("FILTER_DATA"))	
	cerr << " " << _Ustate(0,0) << " " << _Ustate(1,0) << " " << _Ustate(2,0) << " " << _Ustate(3,0) << " " << _UP(0,0) << " " << _UP(1,1) << " " << _UP(2,2) << " " << _UP(3,3) << endl;
}


// this method does a prediction "pinterval" into the future
// using sigma points instead of a standard quaternion prediction
Quat
UnscentedKalmanBase::sigmaPredict(double pinterval, KALMANMODEL mod)
{

 // first compute square root of covariance matrix 
  // use cholesky factorization
  MMatrix sqrtP = choleskyDecomp(_UP);

  // calculate the sigma points
  MMatrix sigma1(_dim, 2 * _dim + 1);
  double coeff = sqrt(_dim + (alpha*alpha * (_dim + kappa) - _dim));
  sigma1.insertColumn(_Ustate, 0);
  for (int i = 1; i < 2*_dim + 1; ++i) 
    if (i < _dim+1) 
     sigma1.insertColumn(_Ustate + (coeff * sqrtP.getColumn(i-1)), i);
    else
     sigma1.insertColumn(_Ustate - (coeff * sqrtP.getColumn(i-_dim-1)), i);
  
   
  // now we must progagate the sigma vectors through the nonlinear equations
  for (int i = 0; i < 2 * _dim + 1; ++i)
    sigma1.insertColumn(_uup->RK4(sigma1.getColumn(i),pinterval, mod), i);	
		
  // compute the predicted quaternion
  MMatrix upstate(_dim,1);
  double lambda = alpha*alpha*(_dim + kappa) - _dim;
  for (int i = 0; i < 2 * _dim + 1; ++i) 
    if (i == 0)
      upstate += (lambda / (lambda + _dim)) * sigma1.getColumn(i);
    else
      upstate += (1.0 / (2.0 * (lambda + _dim))) * sigma1.getColumn(i);

   return normalize(Quat(upstate(0,0), upstate(1,0), upstate(2,0), upstate(3,0)));
}

