#include "types.H"
#include "mmae.H"
#include "mmae3model.H"
#include "mmae3modelc.H"
#include "mmaemultip.H"

// initialize all of the different mmae's
MMAEPredictor::MMAEPredictor()
{
 _mod3 = new MMAE3Model();
 _mod3c = new MMAE3ModelComp();
 _mparam = new MMAEMultiParam();
}


// destructor
MMAEPredictor::~MMAEPredictor()
{
  delete _mod3;
  delete _mod3c;
  delete _mparam;
}

void
MMAEPredictor::prepare() 
{
  // get some things we need from the bundle and send them to
  // the appropriate filter
  double spd = (double)_paramb->getMMAESpecD();
  double delta = 1.0 / _paramb->getCurSRate();
  double mnoisep = _paramb->getPosVar();
  double mnoiseq = _paramb->getQuatVar();
  double sep = (double)_paramb->getSPSep();
  CombinerType cmb = _paramb->getCombiner();

  switch (_paramb->getMMAEType()) {
    case tMMAE3MOD:
      _mod3->initialize(spd, delta, mnoisep, mnoiseq, cmb);
      break;
    case tMMAE3MODComp:
      _mod3c->initialize(spd, delta, mnoisep, mnoiseq, cmb);
      break;
    case tMMAEMulParam: 
      _mparam->initialize(spd, sep, delta, mnoisep, mnoiseq, cmb);
    default: 
    cerr << "No MMAE Predictor Selected!!" << endl;
  }

}

Pose *
MMAEPredictor::predict(Pose *userpos, double pinterval)
{
  switch (_paramb->getMMAEType()) {
    case tMMAE3MOD:
      return _mod3->filter(userpos, pinterval);
      break;
    case tMMAE3MODComp:
      return _mod3c->filter(userpos, pinterval);
      break;
    case tMMAEMulParam:
      return _mparam->filter(userpos, pinterval);
      break;
    default:
      cerr << "No Predictor Selected!!!" << endl;
      // for testing, just make everythin constant
      Pose *tmp = new Pose(userpos->_tstamp+pinterval, Vec3d(0.0, 0.0, 0.0), Quat(0.0,0.0,0.0,1.0));
      return tmp;
  }
}
