// this class contains all of the computational methods for
// analyzing and quantifying motion signal datasets
// It will compute and display results.

#ifndef SANALYZER_H
#define SANALYZER_H

#include "types.H"
#include "gauss.H"
#include "list.H"


class GlobalWidget;
class Pose;
class Vec3d;
class Quat;

// simple class for holding autocorrelation data points
class ACFPts
{
 public:
   ACFPts(Vec3d pos, Quat ori) : _pos(pos), _orient(ori)
   { }
   virtual ~ACFPts()  { }

   ACFPts() : _pos(0.0,0.0,0.0), _orient(0.0,0.0,0.0,1.0)  
   { }
  
   void setZeros() { _pos[0] = _pos[1] = _pos[2] = 0.0;
                     _orient.V()[0] = _orient.V()[1] = _orient.V()[2] = 0.0;
                     _orient.S() = 0.0; }
 
 public:
   Vec3d _pos;
   Quat  _orient;
};


class ACFPtList
{
 public:
   ACFPtList(int size) : _ptlist(size)
   { }
   virtual ~ACFPtList() { }

 public:
   FIGarray<ACFPts*> _ptlist;
};   


class SignalAnalyzer
{
 public:
   SignalAnalyzer(GlobalWidget *gbw);
   virtual ~SignalAnalyzer();

   void prepare();
   
   // analysis functions
   void computeSACF(int lag, int maxlag = 0);
   void computeACFWindows(int initwin);  
   void computeBasicSignalMetrics();
   void computeBasicVelMetrics();
   void computeAndDisplayPSD();
   void computeAndDisplayCrPt();
  
   Pose * addNoise(Pose *cpos, int srate, double noisev);
 
   // display functions 
   void displaySACFResults();
   void displayWinACFResults();
   void displayBasicSignalMetrics();
   void displayBasicVelMetrics();

   void plot(const char *varname, const char *subplot, 
             const char *plotcmd, int numl, double *data,
             const char *title, const char *xlabel, 
             const char *ylabel);

   void plotWin(double *data, int num);
   void plotBounds(const char *title);


   // individual analysis functions for model building
   double computePosVar(FIGarray<Pose*> plist);
   double computeRotVar(FIGarray<Pose*> &plist);
   double computePosSpd(FIGarray<Pose*> &plist);
   double computeRotSpd(FIGarray<Pose*> &plist);
   double computeDPosVar(FIGarray<Pose*> &plist);
   double computeDRotVar(FIGarray<Pose*> &plist);
  
 protected:
   GlobalWidget *_gbw;
   GaussWhiteNoise *_posnoise;
   GaussWhiteNoise *_quatnoise;

   FIGarray<Pose*>   _tmprec;  

   FIGarray<ACFPts*> _acf;
   FIGarray<ACFPtList*> _acflists;

   // used for basic signal metrics
   ACFPts _tpower;
   ACFPts _apower;
   ACFPts _var;
   ACFPts _rms;

   // used for basci velocity-based metrics
   ACFPts _avgabsvel;
   ACFPts _velvar;

   int         _srate;
   double      _noisev;


};

#endif
